searchNodes=[{"doc":"Entrypoint module for Ecto related function with Loupe. Ideally, this module should be completely decoupled from any Repo logic and leave that to the app's Repo","ref":"Loupe.Ecto.html","title":"Loupe.Ecto","type":"module"},{"doc":"Builds an Ecto query from either an AST or a string. It requires an implementation of the Loupe.Ecto.Definition behaviour and supports assigns as a third parameter.","ref":"Loupe.Ecto.html#build_query/3","title":"Loupe.Ecto.build_query/3","type":"function"},{"doc":"","ref":"Loupe.Ecto.html#t:ast/0","title":"Loupe.Ecto.ast/0","type":"type"},{"doc":"The context is the structure that goes through the query building process. It includes the user assigns which are passed down below in the ecto definition, but also includes a couple fields to make sure that query info are validated beforehand. Ecto schema validations Some things are validated before executing the query to avoid crashes during the query execution. These validations includs: Validate that the fetched schema exists and a valid schema. Validate that the queried field are valid on the attached schema and allowed. It also extract the query bindings which are the queries associations and validates them. This is meant to allow someone to use posts.comments.user.name in a query and directly join the posts's comments and comments's user to filter its name. Binding naming Ecto requires named bindings to be atom (which makes total sense). To avoid generating atoms at runtimes in the system, any non-reserverd keyword is cast as string when generating the AST. When building the Ecto query, the join statements uses bindings from a predefined set that you can find a the bottom of this file under the @binding_keys module attribute.","ref":"Loupe.Ecto.Context.html","title":"Loupe.Ecto.Context","type":"module"},{"doc":"Instanciates a context with an implementation and some assigns. The assigns will be passed down to the implementation during to query building process to alter the definition dynamically.","ref":"Loupe.Ecto.Context.html#new/2","title":"Loupe.Ecto.Context.new/2","type":"function"},{"doc":"Puts bindings in the context validation that bindings are either valid fields or associations.","ref":"Loupe.Ecto.Context.html#put_bindings/2","title":"Loupe.Ecto.Context.put_bindings/2","type":"function"},{"doc":"Puts root schema in the context validating it's a valid Ecto schema","ref":"Loupe.Ecto.Context.html#put_root_schema/2","title":"Loupe.Ecto.Context.put_root_schema/2","type":"function"},{"doc":"Checks if a given schema field is allowed","ref":"Loupe.Ecto.Context.html#schema_field_allowed?/3","title":"Loupe.Ecto.Context.schema_field_allowed?/3","type":"function"},{"doc":"Gets implementation schema fields","ref":"Loupe.Ecto.Context.html#schema_fields/2","title":"Loupe.Ecto.Context.schema_fields/2","type":"function"},{"doc":"Gets implementation schemas","ref":"Loupe.Ecto.Context.html#schemas/1","title":"Loupe.Ecto.Context.schemas/1","type":"function"},{"doc":"Sorts binding in order to be join in order","ref":"Loupe.Ecto.Context.html#sorted_bindings/1","title":"Loupe.Ecto.Context.sorted_bindings/1","type":"function"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:assigns/0","title":"Loupe.Ecto.Context.assigns/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:binding_path/0","title":"Loupe.Ecto.Context.binding_path/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:bindings/0","title":"Loupe.Ecto.Context.bindings/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:implementation/0","title":"Loupe.Ecto.Context.implementation/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:schema/0","title":"Loupe.Ecto.Context.schema/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:schemas/0","title":"Loupe.Ecto.Context.schemas/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:t/0","title":"Loupe.Ecto.Context.t/0","type":"type"},{"doc":"Behaviour for defining an Ecto definition. This behaviours defines what the query builder can use to build the query. All function receives the context assigns, meaning that you could add a user's role to the assigns and filter down available schemas for this particular role (example, allow Users query only for Admins) Exmaple In the following example, the Ecto definition prevents non-admin to query User and only allow the :name field for the Game schema. Meaning that non-admin cannot query for other fields than :name when querying games. defmodule MyApp.Loupe.EctoDefinition do @behaviour Loupe.Ecto.Definition @impl Loupe.Ecto.Definition def schemas ( %{ role : &quot;admin&quot; } ) , do : %{ &quot;User&quot; =&gt; MyApp.User , &quot;Game&quot; =&gt; MyApp.Game } def schemas ( _ ) , do : %{ &quot;Game&quot; =&gt; MyApp.Game } @impl Loupe.Ecto.Definition def schema_fields ( Game , %{ role : role } ) when role != &quot;admin&quot; , do : [ :name ] def schema_fields ( _ , _ ) , do : :all @impl Loupe.Ecto.Definition def scope_schemas ( Game , %{ role : role } ) when role != &quot;admin&quot; do Game . not_deleted ( ) end def scope_schemas ( schema , _ ) , do : schema end","ref":"Loupe.Ecto.Definition.html","title":"Loupe.Ecto.Definition","type":"behaviour"},{"doc":"Gets available field for a schema","ref":"Loupe.Ecto.Definition.html#c:schema_fields/2","title":"Loupe.Ecto.Definition.schema_fields/2","type":"callback"},{"doc":"Gets available schemas","ref":"Loupe.Ecto.Definition.html#c:schemas/1","title":"Loupe.Ecto.Definition.schemas/1","type":"callback"},{"doc":"Scopes schema query builder","ref":"Loupe.Ecto.Definition.html#c:scope_schema/2","title":"Loupe.Ecto.Definition.scope_schema/2","type":"callback"},{"doc":"Loupe's compiler entrypoint.","ref":"Loupe.Language.html","title":"Loupe.Language","type":"module"},{"doc":"Compiles a query to AST","ref":"Loupe.Language.html#compile/1","title":"Loupe.Language.compile/1","type":"function"},{"doc":"","ref":"Loupe.Language.html#t:compile_error/0","title":"Loupe.Language.compile_error/0","type":"type"},{"doc":"Extracted AST structure from a get query. It uses a basic syntax like get [ quantifier? ] [ schema ] where [ predicates ] The quantifier is used to limit the queries result but can be ommited defaulting to 1 . It supports the following: Positive integer; 1 , 2 , 10 etc... Range: 10..20 , it limits the query to 10 records offsetting to the 10th record all : Returns all the record matching The schema needs to be an idenfifier (non-quoted alphanumeric) that matches the definition's schemas/1 function. The predicates are combination of boolean operators and operand for validation. See the module's type for every support operators but it can basically be a syntax like get 5 User where ( name = &quot;John Doe&quot; ) or ( age &gt; 18 )","ref":"Loupe.Language.GetAst.html","title":"Loupe.Language.GetAst","type":"module"},{"doc":"Extracts bindings of an AST","ref":"Loupe.Language.GetAst.html#bindings/1","title":"Loupe.Language.GetAst.bindings/1","type":"function"},{"doc":"","ref":"Loupe.Language.GetAst.html#is_boolean_operator/1","title":"Loupe.Language.GetAst.is_boolean_operator/1","type":"macro"},{"doc":"","ref":"Loupe.Language.GetAst.html#is_literal/1","title":"Loupe.Language.GetAst.is_literal/1","type":"macro"},{"doc":"","ref":"Loupe.Language.GetAst.html#is_operand/1","title":"Loupe.Language.GetAst.is_operand/1","type":"macro"},{"doc":"","ref":"Loupe.Language.GetAst.html#is_reserved_keyword/1","title":"Loupe.Language.GetAst.is_reserved_keyword/1","type":"macro"},{"doc":"","ref":"Loupe.Language.GetAst.html#is_text_operand/1","title":"Loupe.Language.GetAst.is_text_operand/1","type":"macro"},{"doc":"Instanciates the AST","ref":"Loupe.Language.GetAst.html#new/3","title":"Loupe.Language.GetAst.new/3","type":"function"},{"doc":"Composed bidings from nested querying","ref":"Loupe.Language.GetAst.html#t:binding/0","title":"Loupe.Language.GetAst.binding/0","type":"type"},{"doc":"Valid boolean operators","ref":"Loupe.Language.GetAst.html#t:boolean_operator/0","title":"Loupe.Language.GetAst.boolean_operator/0","type":"type"},{"doc":"Literial values usable in comparison","ref":"Loupe.Language.GetAst.html#t:literal/0","title":"Loupe.Language.GetAst.literal/0","type":"type"},{"doc":"Valid comparison operands","ref":"Loupe.Language.GetAst.html#t:operand/0","title":"Loupe.Language.GetAst.operand/0","type":"type"},{"doc":"Validation composed predicates","ref":"Loupe.Language.GetAst.html#t:predicate/0","title":"Loupe.Language.GetAst.predicate/0","type":"type"},{"doc":"Query quantifier to limit the query result count","ref":"Loupe.Language.GetAst.html#t:quantifier/0","title":"Loupe.Language.GetAst.quantifier/0","type":"type"},{"doc":"Range from one value to another","ref":"Loupe.Language.GetAst.html#t:range/0","title":"Loupe.Language.GetAst.range/0","type":"type"},{"doc":"Reserved keywords","ref":"Loupe.Language.GetAst.html#t:reserved_keyword/0","title":"Loupe.Language.GetAst.reserved_keyword/0","type":"type"},{"doc":"","ref":"Loupe.Language.GetAst.html#t:t/0","title":"Loupe.Language.GetAst.t/0","type":"type"},{"doc":"An error that occured in the lexer's step","ref":"Loupe.Errors.LexerError.html","title":"Loupe.Errors.LexerError","type":"exception"},{"doc":"","ref":"Loupe.Errors.LexerError.html#t:t/0","title":"Loupe.Errors.LexerError.t/0","type":"type"},{"doc":"An error that occured in the parser's step","ref":"Loupe.Errors.ParserError.html","title":"Loupe.Errors.ParserError","type":"exception"},{"doc":"","ref":"Loupe.Errors.ParserError.html#t:t/0","title":"Loupe.Errors.ParserError.t/0","type":"type"},{"doc":"Loupe is query language for Ecto schema inspection in a safe and configurable manner.","ref":"readme.html","title":"Loupe","type":"extras"},{"doc":"If available in Hex , the package can be installed by adding loupe to your list of dependencies in mix.exs : def deps do [ { :loupe , &quot;~&gt; 0.1.0&quot; } ] end Documentation can be generated with ExDoc and published on HexDocs . Once published, the docs can be found at https://hexdocs.pm/loupe .","ref":"readme.html#installation","title":"Loupe - Installation","type":"extras"},{"doc":"The basic syntax has the following format get [ quantifier? ] [ schema ] where [ predicates ] quantifier is how many records you want. You can provide a positive integer ( 1 , 2 , 3 ...), a range ( 1..10 , 10..20 , 50..100 ) or all . schema needs to be an alphanumeric indentifier that you registered in the Definition (See Ecto Usage for exmaple). predicates needs to be a combinaison or operators and boolean operators. Cool stuff You can use k and m quantifiers for numbers. Writing get all User where money &gt; 100k translates to get all User where money &gt; 100000 . Operators The are a couple of basic operators like &lt; , &gt; , &lt;= , &gt;= , = , != . But also some textual operators: in is used with lists, like age in [18, 21] like is used with strings and automatically wraps in % . You can also use the keyword :empty as a null checker like age :empty . Textual operators and :empty can be prefixed with not to negate the expression: not like , not in , age not :empty . For boolean, the binding can be provided as is and prefixed by not for false. Example where active or where not enabled . Boolean Operators So far, the syntax supprts and and or and use parenthese to scope the expressions.","ref":"readme.html#syntax","title":"Loupe - Syntax","type":"extras"},{"doc":"Create a Definition module The Definition module is necessary for Loupe to work with your Ecto schema. In this module you define the schemas that are allowed to be queried and the fields that are permitted for querying. All callbacks accepts a last argument called &quot;assigns&quot;. The assigns are provided to you when evaluating the query allowing you to alter the defition. You could, for instance, add a user's role to the assign and use that role to filter out the allowed schemas so that only admins can query Users. defmodule MyApp.Loupe.Definition do @moduledoc &quot;&quot;&quot; Example Ecto definition for the modules defined above. &quot;&quot;&quot; @behaviour Loupe.Ecto.Definition @schemas %{ &quot;Post&quot; =&gt; Post , &quot;User&quot; =&gt; User , &quot;Role&quot; =&gt; Role } @impl Loupe.Ecto.Definition def schemas ( %{ role : &quot;admin&quot; } ) , do : @schemas def schemas ( _ ) , do : Map . take ( @schemas , [ &quot;Post&quot; , &quot;User&quot; ] ) @impl Loupe.Ecto.Definition def schema_fields ( _ , %{ role : &quot;admin&quot; } ) , do : :all def schema_fields ( Post , _ ) , do : { :only , [ :title , :body ] } def schema_fields ( User , _ ) , do : { :only , [ :email , :posts ] } def schema_fields ( _ , _ ) , do : :all @impl Loupe.Ecto.Definition def scope_schema ( schema , _ ) , do : schema end Once you have this definition, you can try some queries { :ok , ast } = Loupe.Language . compile ( ~s|get all User where age &gt; 18| ) { :ok , ecto_query } = Loupe.Ecto . build_query ( ast , MyApp.Loupe.Definition , %{ role : &quot;admin&quot; } ) Repo . all ( ecto_query )","ref":"readme.html#ecto-usage","title":"Loupe - Ecto usage","type":"extras"},{"doc":"Here are some things that I would like Loupe to support: Sorting a query, current ideas involves get all User order asc inserted_at get all User where age &gt; 10 ordered asc inserted_at . Support some more complex fields prefixed by ~ (or whatever syntax, inspired by elixir's sigils) like the examples below get all Product where price = ~99.99$ and have that use the Elixir money lib. get all Item where ratio = ~1/4 Replace SQLite in test for Postgres and replace like for ilike in the Ecto Query builder. It has been done like that to avoid messing around with a postgres container only for test but I think it's better to go that route for maximum featureset. Implement a LiveView UI lib that shows the strucutres as expandable. Being able to click on a User's posts to automatically preload all of its nested Posts. Also have &quot;block&quot; UI module where you can simply create a query from dropdowns in a form for non-power user. Make lexer and parser swappable. Right now, you are stuck with the internal structure that I came up with. The idea would be to allow some to swap the syntax for anything they want. For instance, a french team could implement a french query language to give to their normal user.","ref":"readme.html#todo","title":"Loupe - Todo","type":"extras"},{"doc":"You can see the CONTRIBUTING.md file to know more about the contributing guidelines. Pull requests are welcome!","ref":"readme.html#contributing","title":"Loupe - Contributing","type":"extras"}]